/*
    FEARLUS model0-6-4: BitItem.m
    Copyright (C) 1999-2002  Macaulay Institute

    This file is part of FEARLUS model0-6-4, an agent-based model of land use
    change.

    FEARLUS model0-6-4 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-6-4 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the bitItem class. This is just a root class to contain an
unsigned. The bits are stored starting at the least significant bit.

*/

#import "BitItem.h"
#import "FearlusArguments.h"

static void print_binary(FILE *fp, unsigned number, int nbits) {
  int i;

  if(nbits == 0) {
    fprintf(fp, "<Empty BitItem>");
    return;
  }
  for(i = 0; i < nbits; i++) {
    if(number & 1) {
      fputc('1', fp);
    }
    else {
      fputc('0', fp);
    }
    number >>= 1;
  }
}


@implementation BitItem

/*

set_Bits: An accessor function enabling the bits to be set.

*/

-(void)set_Bits: (unsigned)value {
  bits = value;
}

/*

get_Bits: An accessor function returning the bits.

*/

-(unsigned)get_Bits {
  return(bits);
}

/*

set_Bits: Set a particular bit to a specific value

*/

-(void)set_Bit: (int)n to: (int)value {
  unsigned b;

  if((value == 0) || (value == (int)'F') || (value == (int)'f') ||
     (value == (int)'N') || (value == (int)'n')) {
    b = ~(1 << n);	// create a string like 111110111
    bits &= b;				// unset the bit
  }
  else {
    b = 1 << n;		// create a string like 000001000
    bits |= b;				// set the bit
  }
}

/*

get_Bit: Return the value of a particular bit

*/

-(int)get_Bit: (int)n {
  if((bits & (1 << n)) > 0) {
    return(1);
  }
  else {
    return(0);
  }
}

/*

toggle_Bit: Toggle a particular bit (i.e. if it is 0, set it to 1, if 1, set it
to 0).

*/

-(void)toggle_Bit: (int)n {
  bits ^= (1 << n);
}

/*

set_EachBitTo:withProbability: All bits in the BitItem are set to the one's
complement of the first argument to this method (value). Then each bit is
taken in turn, and set to value if a random number between 0 and 1 is less
than prob.

*/

-(void)set_EachBitTo: (int)value withProbability: (double)prob {
  int i;

  if((value == 0) || (value == (int)'F') || (value == (int)'f') ||
     (value == (int)'N') || (value == (int)'n')) {
    value = 0;
    bits = (~0) >> ([self get_NoOfBits] - bitsThatCount);
  }
  else {
    value = 1;
    bits = 0;
  }
  for(i = 0; i < bitsThatCount; i++) {
    if([uniformDblRand getDoubleWithMin: 0.0 withMax: 1.0] < prob) {
      [self set_Bit: i to: value];
    }
  }
}

/*

getSubBitsFrom:to: An accessor function returning a subsection of the bit
item. (inclusive, 0 is the first)

*/

-(unsigned)get_SubBitsFrom: (int)bit1 to: (int)bit2 {
  unsigned subBits, subAND;

  bit2 = [self get_NoOfBits] - bit2;
  subAND = ((~0) >> bit2) & ((~0) << bit1);
  subBits = (bits & subAND) >> bit1;
  return(subBits);
}

/*

set_ToSubBitsOf:from:to: A function setting this bit item to a subsection of
another bit item

*/

-(void)set_ToSubBitsOf: (BitItem *)bits1 from: (int)bit1 to: (int)bit2 {
  bits = [bits1 get_SubBitsFrom: bit1 to: bit2];
  bitsThatCount = (bit2 - bit1 + 1);
}

/*

get_NoOfBits: A function returning the number of bits this data item can hold
-- which is simply sizeof(unsigned) * 8.

*/

-(int)get_NoOfBits {
  return(sizeof(bits) * 8);
}

/*

getMaxBits: A function returning the number of bits the class can store

*/

+(int)get_MaxBits {
  return(sizeof(unsigned) * 8);
}

/*

getBitsThatCount: A function returning the number of bits that actually have
any meaning in this bit item.

*/

-(int)get_BitsThatCount {
  return(bitsThatCount);
}

/*

setBitsThatCount: Some of the bits don't count, because they are not part of
the bit string. That is, enable the bit string to cope with strings that are
not integral multiples of (sizeof(bits) * 8) bits in length

*/

-(void)set_BitsThatCount: (int)value {
  bitsThatCount = value;
}

/*

set_ToCopyOf: A method to make a copy of a bit item this bit item.

*/

-(void)set_ToCopyOf: (BitItem *)bits1 {
  bits = [bits1 get_Bits];
  bitsThatCount = [bits1 get_BitsThatCount];
}

/*

set_ToConcatenationOf: A function taking two bit items and setting this bit
item to be a concatenation of bits1 and bits2. Returned is the number of least
significant bits of bits2 that have not been concatenated.

*/

-(int)set_ToConcatenationOf: (BitItem *)bits1 and: (BitItem *)bits2 {
  int max, n1, n2;
  unsigned b1, b2;

  max = [self get_NoOfBits];
  n1 = [bits1 get_BitsThatCount];
  n2 = [bits2 get_BitsThatCount];
  b1 = [bits1 get_Bits];
  b2 = [bits2 get_Bits];
  if(n1 == max) {		// bits1 uses up all the bits. All the bits in
    bits = b1;			// bits2 are left over. Deal with this one
    bitsThatCount = n1;		// swiftly.
    return(n2);
  }
  b2 <<= n1;			// Shift bits2 left by the no of bits in bits1
  b1 &= ((~0) >> (max - n1));
                                // Clean off any bits in bits1 to be ignored
  bits = b1 | b2;		// Shove the shifted bits2 onto the end of b1
  if((n1 + n2) > max) {
    bitsThatCount = max;
    return((n1 + n2) - max);	// bits1 and bits2 are too big to fit, return
  }				// the number of bits in bits2 that are left
  else {
    bitsThatCount = n1 + n2;
    return 0;
  }
}

/*

set_ToParityOf: This bit item is going to get set to the bitwise
parity of the two bit items in the arguments. If they have different
numbers of bits that count, then this bit item's bits that count will
be the lesser of the two.

*/

-(void)set_ToParityOf: (BitItem *)bits1 and: (BitItem *)bits2 {
  int b;

  bits = ~([bits1 get_Bits] ^ [bits2 get_Bits]);
  // Parity is the one's complement of XOR
  bitsThatCount = [bits1 get_BitsThatCount];
  b = [bits2 get_BitsThatCount];
  if(bitsThatCount > b) {
    bitsThatCount = b;
  }
}

/*

set_ToOROf: Set the bit item of self to bits1 bitwise OR bits2. If
they have different numbers of bits that count, then this object's
bits that count is the lesser.

*/

-(void)set_ToOROf: (BitItem *)bits1 and: (BitItem *)bits2 {
  int b;

  bits = [bits1 get_Bits] | [bits2 get_Bits];
  bitsThatCount = [bits1 get_BitsThatCount];
  b = [bits2 get_BitsThatCount];
  if(bitsThatCount > b) {
    bitsThatCount = b;
  }
}

/*

set_ToXOROf: Set the bit item of self to b1 bitwise XOR b2. If they
have different numbers of bits that count, then this object's bits
that count is the lesser.

*/

-(void)set_ToXOROf: (BitItem *)b1 and: (BitItem *)b2 {
  int b;

  bits = [b1 get_Bits] ^ [b2 get_Bits];
  bitsThatCount = [b1 get_BitsThatCount];
  b = [b2 get_BitsThatCount];
  if(bitsThatCount > b) {
    bitsThatCount = b;
  }
}

/*

get_MatchWith: Return the number of 1s in the result of a parity operation of
self with the argument. The argument must have the same number of bits that
count as self

*/

-(int)get_MatchWith: (BitItem *)anItem {
  unsigned b;
  int match;

  if([anItem get_BitsThatCount] != bitsThatCount) {
    fprintf(stderr,
	    "Attempt to compute match with bit items of unequal length\n");
    return(-1);
  }
  b = bits;
  bits = ~(bits ^ [anItem get_Bits]);
  match = [self count_Ones];
  if([FearlusArguments debugBitstrings]) {
    printf("B: %s: Computed match of BitItems as follows:\n",
	   sel_get_name(_cmd));
    printf("\t   BitItem A: ");
    print_binary(stdout, b, bitsThatCount);
    printf("\n\t   BitItem B: ");
    [anItem print_];
    printf("\tnot(A xor B): ");
    print_binary(stdout, bits, bitsThatCount);
    printf(" -- number of 1s = match = %d\n", match);
  }
  bits = b;
  return(match);
}

/*

set_ToCrossoverOf: Set self's bits to the result of a crossover operation of
bits1 and bits2 at bit number n. The result is a string of length the same as
bits2, consisting of the first n bits of bits1 and the bits n+1 to end of bits2
-- unless the crossover point is after the end of bits2, in which case the
result is a copy of bits1

*/

-(void)set_ToCrossoverOf: (BitItem *)bits1 and: (BitItem *)bits2 at: (int)n {
  unsigned b;

  if(n > [bits2 get_BitsThatCount]) {
    [self set_ToCopyOf: bits1];
    return;
  }
  [self set_ToCopyOf: bits2];	// Make a copy of bits2 in self
  b = (~0) << n;
  bits &= b;			// Wipe out the first n bits of self
  b = ~b;
  // Put the first n bits of bits1 into self
  bits |= (b & [bits1 get_Bits]);
}

/*

mutate_WithProbability: For each bit in turn, mutate it if a random number is
less than the required probability. (Mutation means, in this context, a random
new value for each bit is selected. Half the time, mutation will have no
effect. This means maximum randomness is achieved with a probability equal to
1. The alternative is to forcibly toggle the bits selected for mutation. This
means maximum randomness is at a probability of 0.5, with a probability of 1
corresponding to setting the bit string to its ones complement. This makes less
sense -- well it does to me, anyway.)

*/

-(void)mutate_WithProbability: (double)prob {
  int i, value;

  if(prob == 0.0) return;	// Save time if prob is 0
  for(i = 0; i < bitsThatCount; i++) {
    if([uniformDblRand getDoubleWithMin: 0.0 withMax: 1.0] < prob) {
      value = [uniformIntRand getIntegerWithMin: 0 withMax: 1];
      [self set_Bit: i to: value];
    }
  }
}

/*

count_Ones: Return the number of ones in bits

*/

-(int)count_Ones {
  int n, i;
  unsigned b;

  b = bits;
  if([FearlusArguments debugBitstrings]) {
    printf("Counting the number of 1s in:\n");
    print_binary(stdout, bits, bitsThatCount);
    printf("\n");
  }
  n = 0;
  for(i = 0; i < bitsThatCount; i++) {
    n += b & 1;
    b >>= 1;
    if([FearlusArguments debugBitstrings]) {
      int j;

      for(j = 0; j < i; j++) {
	printf(" ");
      }
      printf("^ tot 1s = %d\n", n);
    }
  }
  return(n);
}

/*

set_ToString: Take in a binary number represented in a string, and set bits to
the value. Return a pointer to any bit of the string not allocated. The string
is assumed to be NULL terminated.

*/

-(char *)set_ToString: (char *)aString {
  char *p;
  int max;
  unsigned b;

  max = [self get_NoOfBits];
  bitsThatCount = 0;
  bits = 0;
  if(strcmp(aString, "") == 0 || strcmp(aString, "E") == 0
     || strcmp(aString, "<EmptyBitString>") == 0) {
    return(aString);
  }
  for(p = aString, b = 1; (bitsThatCount < max) && (*p != '\0');
			  p++, b <<= 1, bitsThatCount++) {
    switch(*p) {
    case '1':
    case 'T':
    case 't':
    case 'Y':
    case 'y':
      bits |= b;
      break;
    case '0':
    case 'F':
    case 'f':
    case 'N':
    case 'n':
      break;
    default:
      fprintf(stderr, "Warning: Unexpected value in binary string %s: %c\n",
	      aString, *p);
    }
  }
  return(p);
}

/*

print_ToFile: Print the bits to a file

*/

-(void)print_ToFile: (FILE *)fp {
  print_binary(fp, bits, bitsThatCount);
}

/*

print_ToString: Print the bits to a string

*/

-(void)print_ToString: (char *)aBuffer {
  int i;
  unsigned b;

  b = bits;
  for(i = 0; i < bitsThatCount; i++) {
    if(b & 1) {
      aBuffer[i] = '1';
    }
    else {
      aBuffer[i] = '0';
    }
    b >>= 1;
  }
  aBuffer[i] = '\0';
}

/*

print_: Print the bits to stdout followed by a \n. 

*/

-(void)print_ {
  print_binary(stdout, bits, bitsThatCount);
  printf("\n");
}

@end
