/*
    FEARLUS model0-6-4: BitString.m
    Copyright (C) 1999-2002  Macaulay Institute

    This file is part of FEARLUS model0-6-4, an agent-based model of land use
    change.

    FEARLUS model0-6-4 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-6-4 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation of the bitString class. This is a class which stores a string of
bits, and provides mechanisms for performing bitwise operations on it. 

*/

#import <objc/objc-api.h>
#import "BitString.h"
#import "BitItem.h"
#import "NumberArray.h"

@implementation BitString

/*

create:setBitCount. An analogue of the create:setCount method of the array
class, but setting the number of bits that will be stored. However, it is also
necessary to populate the array with bitItems.

*/

+create: (id)aZone setBitCount: (unsigned)aBitCount {
  BitString *newBitString;
  BitItem *newItem;
  unsigned size, nBitItems;
  int i;

  size = (unsigned)[BitItem get_MaxBits];
  if(aBitCount == 0) nBitItems = 1;
  else nBitItems = (aBitCount / size) + (((aBitCount % size) > 0) ? 1 : 0);
  // Call the super-class create method
  newBitString = [super create: aZone];
  // Poke in the information
  newBitString->bitCount = aBitCount;
  newBitString->bitsPerItem = size;
  // Create the array
  newBitString->arr = [Array create: aZone setCount: nBitItems];
  // Populate the array with BitItems.
  for(i = 0; i < (int)nBitItems; i++) {
    newItem = [BitItem createBegin: aZone];
    [newItem set_Bits: 0];
    if((i == (int)nBitItems - 1) && ((aBitCount % size) > 0)) {
      // The last item in the array contains the non-integral units of
      // bitsPerItem bits
      [newItem set_BitsThatCount: (int)(aBitCount % size)];
    }
    else if(aBitCount == 0) {
      [newItem set_BitsThatCount: 0];
    }
    else {
      [newItem set_BitsThatCount: (int)size];
    }
    newItem = [newItem createEnd];
    [newBitString->arr atOffset: i put: newItem];
  }
  // Return the created BitString
  return(newBitString);
}

+create: (id)aZone setToString: (char *)aString {
  BitString *newBitString;
  BitItem *newItem;
  unsigned size, nBitItems, aBitCount;
  char *p;
  int i;


  size = (unsigned)[BitItem get_MaxBits];
  aBitCount = (unsigned)strlen(aString);
  if(aBitCount == 0) nBitItems = 1;
  else nBitItems = (aBitCount / size) + (((aBitCount % size) > 0) ? 1 : 0);
  // Call the super-class create method
  newBitString = [super create: aZone];
  // Poke in the information
  newBitString->bitCount = aBitCount;
  newBitString->bitsPerItem = size;
  // Create the array
  newBitString->arr = [Array create: aZone setCount: nBitItems];
  // Populate the array with BitItems set to the string.
  for(i = 0, p = aString; i < (int)nBitItems; i++) {
    newItem = [BitItem createBegin: aZone];
    p = [newItem set_ToString: p];
    newItem = [newItem createEnd];
    [newBitString->arr atOffset: i put: newItem];
  }
  // Return the created BitString
  return(newBitString);
}

+create: (id)aZone setToCopyOf: (BitString *)aBitString {
  BitString *newBitString;
  BitItem *newItem;
  unsigned size, nBitItems, aBitCount;
  int i;

  size = (unsigned)[BitItem get_MaxBits];
  aBitCount = (unsigned)[aBitString getBitCount];
  if(aBitCount == 0) nBitItems = 1;
  else nBitItems = (aBitCount / size) + (((aBitCount % size) > 0) ? 1 : 0);
  newBitString = [super create: aZone];
  // Poke in the information
  newBitString->bitCount = aBitCount;
  newBitString->bitsPerItem = size;
  // Create the array
  newBitString->arr = [Array create: aZone setCount: nBitItems];
  for(i = 0; i < (int)nBitItems; i++) {
    newItem = [BitItem createBegin: aZone];
    [newItem set_ToCopyOf: [aBitString atOffset: i]];
    newItem = [newItem createEnd];
    [newBitString->arr atOffset: i put: newItem];
  }
  // Return the created BitString
  return newBitString;
}
/*

getBitCount -- return the number of bits stored in this bit string

*/

-(unsigned)getBitCount {
  return(bitCount);
}

/*

setBit:to: Set a bit of self to the specified value

*/

-(void)setBit: (unsigned)n to: (int)value {
  int item, bit;

  item = (int)(n / bitsPerItem);
  bit = (int)(n % bitsPerItem);
  [[self atOffset: item] set_Bit: bit to: value];
}

/*

getBit: Return the value of the specified bit

*/

-(int)getBit: (unsigned)n {
  int item, bit;

  item = (int)(n / bitsPerItem);
  bit = (int)(n % bitsPerItem);
  return([[self atOffset: item] get_Bit: bit]);
}

/*

toggleBit: Toggle the specified bit

*/

-(void)toggleBit: (unsigned)n {
  int item, bit;

  item = (int)(n / bitsPerItem);
  bit = (int)(n % bitsPerItem);
  [[self atOffset: item] toggle_Bit: bit];
}

/*

toggleEachBitUsingProbs:

Take each bit in turn, and toggle its value if a random number between 0 and 1
is less than the corresponding probability specified in the number array passed
as argument.

*/

-(void)toggleEachBitUsingProbs: (NumberArray *)probs {
  unsigned i;

  if((unsigned)[probs getCount] != bitCount) {
    fprintf(stderr, "%s -- NumberArray argument has a number of elements (%d) "
	    "not equal to the number of bits in the bit string (%u)\n",
	    sel_get_name(_cmd), [probs getCount], bitCount);
    exit(1);
  }
  for(i = 0; i < bitCount; i++) {
    if([uniformDblRand getDoubleWithMin: 0.0 withMax: 1.0] <
       [probs getDoubleAtOffset: (int)i]) {
      [self toggleBit: i];
    }
  }
}

/*

setEachBitTo:withProbability: Generate a random bitstring where each bit is
set to the specified value with the specified probability.

*/

-(void)setEachBitTo: (int)value withProbability: (double)prob {
  int i;

  for(i = 0; i < (int)[self getCount]; i++) {
    [[self atOffset: i] set_EachBitTo: value withProbability: prob];
  }
}

/*

setToSubBitsOf:from:to: Set self to be a bitstring consisting of the bits in
the bitstring given as argument starting at bit1 and finishing at bit2
(inclusive)

*/

-(void)setToSubBitsOf: (BitString *)bitString from: (unsigned)bit1 
		   to: (unsigned)bit2 
{
  int i, j;
  BitItem *b, *c;
  int leftover;
  int bit1item, bit1bit, bit2item, bit2bit;

  bit1item = bit1 / bitsPerItem;
  bit1bit = bit1 % bitsPerItem;
  bit2item = bit2 / bitsPerItem;
  bit2bit = bit2 % bitsPerItem;
  // Set self's bitcount to be the right size to hold the sub-bits
  [self setBitCount: ((bit2 - bit1) + 1)];
  // If this can all be done within a single bititem, then do that quickly
  // and return
  if(bit1item == bit2item) {
    [[self atOffset: 0] set_ToSubBitsOf: [bitString atOffset: bit1item] 
			from: bit1bit to: bit2bit];
    return;
  }
  // It's not going to fit in a single bititem. Start by setting the first
  // bititem of self to be sub-bits of the first bit item in the bitstring in
  // the argument that is required.
  [[self atOffset: 0] set_ToSubBitsOf:
		      [bitString atOffset: bit1item]
		      from: bit1bit to: bitsPerItem];
  // Create a couple of dummy bit items for temporary storage purposes.
  b = [BitItem createBegin: [self getZone]];
  b = [b createEnd];
  c = [BitItem createBegin: [self getZone]];
  c = [c createEnd];
  // Now do a concatenation loop, putting the beginning of each bit item in
  // the bitstring onto then end of self's current bititem, and putting what's
  // left over into self's next bititem. As a special case, the last bit item
  // from the bitstring in the argument only needs its bits up to bit2bit stuck
  // onto the end of self.
  for(i = 0, j = bit1item + 1; j <= bit2item; i++, j++) {
    // Set c to be whatever has next to be stuck onto the end of self
    if(j < bit2item) {
      [c set_ToCopyOf: [bitString atOffset: j]];
    }
    else {
      [c set_ToSubBitsOf: [bitString atOffset: bit2item] from: 0 to: bit2bit];
    }
    // Use b to store self's current bit item with as much of c stuck onto the
    // end of it as possible, keeping a note of the number of bits of c left
    // over.
    leftover = [b set_ToConcatenationOf: [self atOffset: i] and: c];
    // Pop b into self's current location
    [[self atOffset: i] set_ToCopyOf: b];
    // If there are any bits left over, then stick them into self's next slot.
    if(leftover > 0) {
      [b set_ToSubBitsOf: c from: [c get_BitsThatCount] - leftover
	 to: [c get_BitsThatCount] - 1];
      [[self atOffset: i + 1] set_ToCopyOf: b];
    }
    // If there aren't then make sure the next cell has no bits that count,
    // ready for the next concatenation
    else {
      [[self atOffset: i + 1] set_BitsThatCount: 0];
    }
  }
  [b drop];
  [c drop];
}

/*

setToCopyOf: -- set self to be a copy of the bit string passed as an argument.

*/

-(void)setToCopyOf: (BitString *)bitString {
  int i;

  [self setBitCount: [bitString getBitCount]];
  for(i = 0; i < (int)[bitString getCount]; i++) {
    [[self atOffset: i] set_ToCopyOf: [bitString atOffset: i]];
  }
}

/*

setToConcatenationOf:and: -- set self to be a concatenation of the two bit
strings in the arguments.

*/

-(void)setToConcatenationOf: (BitString *)bitString1
			and: (BitString *)bitString2
{
  int i, j;
  BitItem *b;
  int leftover;

  // Deal with zero length bit strings first, quickly
  if([bitString1 getBitCount] == 0) {
    if([bitString2 getBitCount] == 0) {
      [self setBitCount: 0];
      return;
    }
    [self setToCopyOf: bitString2];
    return;
  }
  else if([bitString2 getBitCount] == 0) {
    [self setToCopyOf: bitString1];
    return;
  }
  // Make sure the array is the right size to store the concatenated bits
  [self setBitCount: ([bitString1 getBitCount] + [bitString2 getBitCount])];
  // Make a copy of bitString1 in self
  for(i = 0; i < (int)[bitString1 getCount]; i++) {
    [[self atOffset: i] set_ToCopyOf: [bitString1 atOffset: i]];
  }
  // Set up a dummy bitItem
  b = [BitItem createBegin: [self getZone]];
  b = [b createEnd];
  // Concatenate bitString2 to self
  for(i--, j = 0; j < (int)[bitString2 getCount]; j++, i++) {
    // Store a concatenation of the last element of self with the next element
    // of bitString2 in the dummy bitItem
    leftover = [b set_ToConcatenationOf: [self atOffset: i] and:
		  [bitString2 atOffset: j]];
    // Set the last element of self to the dummy bitItem which now stores the
    // concatenation
    [[self atOffset: i] set_ToCopyOf: b];
    // If there are any bits left over from the concatenation which wouldn't
    // fit in self's last bitItem
    if(leftover > 0) {
      // Set the dummy bitItem to store the bits that were left over
      [b set_ToSubBitsOf: [bitString2 atOffset: j] from:
	 [[bitString2 atOffset: j] get_BitsThatCount] - leftover to:
	 [[bitString2 atOffset: j] get_BitsThatCount] - 1];
      // Copy the bits that were left over into the next element of self. Now
      // we might have a little worry that an out-of-bounds call is made to
      // self here, but let's reassure ourselves. Self has been set to the
      // right size for the concatenation at the beginning of this method.
      // If we've just filled the last element of self, then there won't be
      // any bits left over, will there? So we won't ever be trying to fill
      // an out-of-bounds element of self in this method.
      [[self atOffset: i + 1] set_ToCopyOf: b];
    }
    else if(!(j == (int)[bitString2 getCount] - 1)) {
      // If there aren't any bits left over, then make sure self's next bit
      // item has no bits that count, ready to be concatenated into. But of
      // course, we don't want to do this if this is the last time we'll be
      // going through the loop, because there won't be any next bit item.
      [[self atOffset: i + 1] set_BitsThatCount: 0];
    }
  }
  [b drop];			// Rid ourselves of the dummy element.
}

/*

setToParityOf:and: -- set self to be the result of a bitwise parity
operation on bitString1 and bitString2. Self's size will be that of
the smaller of the two bitStrings.

*/

-(void)setToParityOf: (BitString *)bitString1 and: (BitString *)bitString2 {
  unsigned bc;
  int i;

  // set self's size to be the smaller of the two bitstrings in the arguments
  bc = ([bitString1 getBitCount] < [bitString2 getBitCount]) ?
    [bitString1 getBitCount] : [bitString2 getBitCount];
  [self setBitCount: bc];
  // set self to be the result of the parity operation
  for(i = 0; i < (int)[self getCount]; i++) {
    [[self atOffset: i] set_ToParityOf: [bitString1 atOffset: i] and:
			[bitString2 atOffset: i]];
  }
}

/*

setToOROf:and: -- set self to be the result of a bitwise OR operation
on bitString1 and bitString2. Self's size will be that of the smaller
of the two bitstrings.

*/

-(void)setToOROf: (BitString *)bitString1 and: (BitString *)bitString2 {
  unsigned bc;
  int i;

  // set self's size to be the smaller of the two bitstrings in the arguments
  bc = ([bitString1 getBitCount] < [bitString2 getBitCount]) ?
    [bitString1 getBitCount] : [bitString2 getBitCount];
  [self setBitCount: bc];
  // set self to be the result of the OR operation
  for(i = 0; i < (int)[self getCount]; i++) {
    [[self atOffset: i] set_ToOROf: [bitString1 atOffset: i] and:
			[bitString2 atOffset: i]];
  }
}

/*

setToXOROf:and: -- set self to be the result of a bitwise XOR
operation on bs1 and bs2.  Self's size will be that of the smaller of
the two bitstrings.

*/

-(void)setToXOROf: (BitString *)bs1 and: (BitString *)bs2 {
  unsigned bc;
  int i;

  // set self's size to be the smaller of the two bitstrings in the arguments
  bc = ([bs1 getBitCount] < [bs2 getBitCount]) ?
    [bs1 getBitCount] : [bs2 getBitCount];
  [self setBitCount: bc];
  // set self to be the result of the XOR operation
  for(i = 0; i < (int)[self getCount]; i++) {
    [[self atOffset: i] set_ToXOROf: [bs1 atOffset: i] and:
			[bs2 atOffset: i]];
  }
}

/*

setToCrossoverOf:and:at: set self to be the result of a Crossover operation on
bitString1 and bitString2. Self's size will be that of bitString2, unless the
crossover point is after the end of bitString2, in which case self will just be
a copy of bitString1.

*/

-(void)setToCrossoverOf: (BitString *)bitString1 and: (BitString *)bitString2
		     at: (unsigned)bit
{
  int i;
  unsigned crossoverItem, crossoverBit;

  if(bit > [bitString2 getBitCount]) {
    [self setToCopyOf: bitString1];
    return;
  }
  [self setBitCount: [bitString2 getBitCount]];
  crossoverItem = bit / bitsPerItem;
  crossoverBit = bit % bitsPerItem;
  if(crossoverBit != 0) {
    crossoverItem++;
  }
  for(i = 0; i < (int)crossoverItem; i++) {
    [[self atOffset: i] set_ToCopyOf: [bitString1 atOffset: i]];
  }
  if(crossoverBit != 0) {
    [[self atOffset: crossoverItem] set_ToCrossoverOf:
				    [bitString1 atOffset: crossoverItem] and:
				    [bitString2 atOffset: crossoverItem]
				    at: crossoverBit];
  }
  else {
    [[self atOffset: crossoverItem] set_ToCopyOf:
				    [bitString2 atOffset: crossoverItem]];
  }
  for(i = crossoverItem + 1; i < (int)[bitString2 getCount]; i++) {
    [[self atOffset: i] set_ToCopyOf: [bitString2 atOffset: i]];
  }
}

/*

mutateWithProbability: Use the probability to determine for each bit whether or
not it should be mutated. (Mutation means, in this context, a random new value
for each bit is selected. Half the time, mutation will have no effect. This
means maximum randomness is achieved with a probability equal to 1. The
alternative is to forcibly toggle the bits selected for mutation. This means
maximum randomness is at a probability of 0.5, with a probability of 1
corresponding to setting the bit string to its ones complement.)

*/

-(void)mutateWithProbability: (double)prob {
  int i;

  for(i = 0; i < (int)[self getCount]; i++) {
    [[self atOffset: i] mutate_WithProbability: prob];
  }
}

/*

getMatchWith -- count the number of ones in a parity operation of self with
the argument. Both must have the same number of bits.

*/

-(unsigned)getMatchWith: (BitString *)anItem {
  unsigned match;
  int i;

  if(bitCount != [anItem getBitCount]) {
    fprintf(stderr, "Attempt to compute match between bit strings of unequal "
	    "length: self %u, arg %u\n", bitCount, [anItem getBitCount]);
    return(0);
  }
  match = (0);
  for(i = 0; i < (int)[arr getCount]; i++) {
    match += (unsigned)[[self atOffset: i] get_MatchWith:
						[anItem atOffset: i]];
  }
  return match;
}

/*

countOnes -- count the number of ones in the binary string

*/

-(unsigned)countOnes {
  unsigned oneCount;
  int i;

  oneCount = 0;
  for(i = 0; i < (int)[arr getCount]; i++) {
    oneCount += (unsigned)[[self atOffset: i] count_Ones];
  }
  return(oneCount);
}

/*

setToString -- set the bitString to the character strings of 1s and 0s passed
in the argument

*/

-(void)setToString: (char *)aString {
  char *p;
  int i;

  if(strcmp(aString, "") == 0 || strcmp(aString, "E") == 0
     || strcmp(aString, "<Empty BitString>") == 0) {
    [self setBitCount: 0];
    return;
  }
  [self setBitCount: (unsigned)strlen(aString)];
  for(i = 0, p = aString; i < (int)[arr getCount]; i++) {
    p = [[self atOffset: i] set_ToString: p];
  }
}

/*

setBitCount -- An analogue of the setCount method of the Array class, but
(re-)setting the number of bits the bit string will store. BitItem elements 
that are not needed must get destroyed, and new BitItem elements created, as
appropriate for the re-sizing operation that is taking place.

*/

-(void)setBitCount: (unsigned)aBitCount {
  unsigned nBitItems, oldNBitItems;
  BitItem *newItem;
  int i;

  // Calculate the new number of bitItems that will be needed to store this
  // bitstring
  if(aBitCount == 0) {
    nBitItems = 1;
  }
  else {
    nBitItems = (aBitCount / bitsPerItem) +
      (((aBitCount % bitsPerItem) > 0) ? 1 : 0);
  }
  // Keep a note of the old number of bitItems
  oldNBitItems = [arr getCount];
  // Send a message to the current last element to tell it to set its number
  // of bits that count to the maximum.
  if(oldNBitItems > 0) {
    [[arr atOffset: ((int)oldNBitItems - 1)] set_BitsThatCount: bitsPerItem];
  }
  bitCount = aBitCount;			// Set the bitCount to the new value
  if(nBitItems != oldNBitItems) {
    // If the number of bitItems has shrunk or increased, we must:
    // (a) Change the size of the array
    // (b) Create new bitItems to populate new members of the array
    // (c) Ensure that all bitItems are told to store the correct number of
    //     bits.
    // (d) Drop any bitItems that need to be deleted for array re-sizing
    for(i = (int)nBitItems; i < (int)oldNBitItems; i++) {
      [[arr atOffset: i] drop];	// (d)
    }
    [arr setCount: (int)nBitItems];	// (a) setCount is inherited from Array
    for(i = (int)oldNBitItems; i < (int)nBitItems; i++) {
      // (b) Populate any newly allocated array space with new bit items
      newItem = [BitItem createBegin: [self getZone]];	// Create the new item
      [newItem set_Bits: 0];		// Initialise it to 000000....
      [newItem set_BitsThatCount: (int)bitsPerItem];
      [arr atOffset: i put: newItem];
    }
  }
  // (c) Send a message to the new last element to tell it to set its number of
  // bits that count to whatever is left over
  if(bitCount % bitsPerItem > 0) {
    [[self atOffset: ((int)nBitItems - 1)] set_BitsThatCount:
					   (int)(bitCount % bitsPerItem)];
  }
  else if(bitCount == 0) {
    [[self atOffset: 0] set_BitsThatCount: 0];
  }
}


/*

printToFile -- send a printout of the bitString to a file

*/

-(void)printToFile: (FILE *)fp {
  int i;

  if(bitCount == 0) {
    fprintf(fp, "<Empty BitString>");
    return;
  }
  for(i = 0; i < (int)[arr getCount]; i++) {
    [[self atOffset: i] print_ToFile: fp];
  }
}

/*

printToString -- store a printable bitString in a char buffer

*/

-(void)printToString: (char *)aBuffer {
  int i;
  char *p;

  for(p = aBuffer, i = 0; i < (int)[arr getCount]; i++) {
    [[self atOffset: i] print_ToString: p];
    p += [[self atOffset: i] get_BitsThatCount];
  }
}

/*

print -- send a printout of the bitString to stdout, followed by a \n

*/

-print {
  [self printToFile: stdout];
  printf("\n");
  return self;
}

/*

drop -- override the drop method to include an action to drop all
the members of the array first

*/

-(void)drop {
  int i;

  for(i = 0; i < (int)[arr getCount]; i++) {
    [[arr atOffset: i] drop];
  }
  [super drop];
}

/*

atOffset -- provide an atOffset method that only returns objects of
type bitItem.

*/

-(BitItem *)atOffset: (unsigned)i {
  return([arr atOffset: i]);
}

/*

getCount -- provide an accessor for the Array class getCount method

*/

-(unsigned)getCount {
  return([arr getCount]);
}

@end
