/*
    FEARLUS model0-6-4: Double.m
    Copyright (C) 1999-2002  Macaulay Institute

    This file is part of FEARLUS model0-6-4, an agent-based model of land use
    change.

    FEARLUS model0-6-4 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-6-4 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/

/*

Implementation for the Double object. This uses the ieeefp library, if
available.

*/

#import <ieeefp.h>
#import "Double.h"
#import <stdio.h>
#import <errno.h>

/* Class variables */

static double epsilon = 0.0;

static fp_except fpe_flags = 0
#if defined(FP_X_INV)
  | FP_X_INV
#endif
#if defined(FP_X_DNML)
  | FP_X_DNML
#endif
#if defined(FP_X_DZ)
  | FP_X_DZ
#endif
#if defined(FP_X_OFL)
  | FP_X_OFL
#endif
#if defined(FP_X_UFL)
  | FP_X_UFL
#endif
#if defined(FP_X_IMP)
  | FP_X_IMP
#endif
;

#if defined(FP_X_INV) || defined(FP_X_DNML) || defined(FP_X_DZ) \
  || defined(FP_X_OFL) || defined(FP_X_UFL) || defined(FP_X_IMP)
static BOOL fpe = YES;
#else
static BOOL fpe = NO;
#endif


/* Functions */

static void set_fp_flags() {
  fpsetmask(fpe_flags);
}

static void unset_fp_flags() {
  fpsetmask(0);
}

/* Implementation */

@implementation Double

+new {
  return [self new: 0.0];
}

+new: (double)value {
  Double *obj = [super new];

  [obj set: value];

  return obj;
}

+setEpsilon: (double)value {
  epsilon = value;
  return self;
}

+(double)getEpsilon {
  return epsilon;
}

+setFPE: (BOOL)value {
  fpe = value;
  if(value == YES && fpe_flags == 0) {
    fprintf(stderr, "Not able to trap any recognised floating point "
	    "exceptions on the machine this was compiled on\n");
    abort();
  }
  return self;
}

+(BOOL)getFPE {
  return fpe;
}

+(const char **)getTrappableExceptions {
  static const char *exceptxt[] = {
#if defined(FP_X_INV)
    "invalid operation",
#endif
#if defined(FP_X_DNML)
    "denormalization",
#endif
#if defined(FP_X_DZ)
    "divide-by-zero",
#endif
#if defined(FP_X_OFL)
    "overflow",
#endif
#if defined(FP_X_UFL)
    "underflow",
#endif
#if defined(FP_X_IMP)
    "loss of precision",
#endif
    ""
  }
  return exceptxt;
}

+(const char **)getTrappedExceptions {
  static const char *nonetxt[] = {
    "no floating point exceptions currently trapped",
    ""
  }

  if(fpe) {
    return [self getTrappableExceptions];
  }
  else {
    return nonetxt;
  }
}


-copy {
  return [Double new: d];
}


-set: (double)value {
  if(fpe) fpsetmask(fpe_flags);
  d = value;
  if(fpe) fpsetmask(0);
  return self;
}

-(double)get {
  return d;
}

-add: (Double *)value {
  if(fpe) fpsetmask(fpe_flags);
  d += [value get];
  if(fpe) fpsetmask(0);
  return self;
}

-sub: (Double *)value {
  if(fpe) fpsetmask(fpe_flags);
  d -= [value get];
  if(fpe) fpsetmask(0);
  return self;
}

-mul: (Double *)value {
  if(fpe) fpsetmask(fpe_flags);
  d *= [value get];
  if(fpe) fpsetmask(0);
  return self;
}

-div: (Double *)value {
  if(fpe) fpsetmask(fpe_flags);
  d /= [value get];
  if(fpe) fpsetmask(0);
  return self;
}


-(BOOL)eq: (Double *)value {
  BOOL eq;
  double v = [value get];
  if(fpe) fpsetmask(fpe_flags);
  eq = (d <= v + epsilon && d >= v - epsilon) ? YES : NO;
  if(fpe) fpsetmask(0);
  return eq;
}

-(BOOL)ne: (Double *)value {
  BOOL ne;
  double v = [value get];
  if(fpe) fpsetmask(fpe_flags);
  ne = (d > v + epsilon || d < v - epsilon) ? YES : NO; 
  if(fpe) fpsetmask(0);
  return ne;
}

-(BOOL)gt: (Double *)value {
  BOOL gt;
  if(fpe) fpsetmask(fpe_flags);
  gt = (d > [value get] + epsilon) ? YES : NO;
  if(fpe) fpsetmask(0);
  return gt;
}

-(BOOL)le: (Double *)value {
  BOOL le;
  if(fpe) fpsetmask(fpe_flags);
  le = (d <= [value get] - epsilon) ? YES : NO;
  if(fpe) fpsetmask(0);
  return le;
}

-(BOOL)ge: (Double *)value {
  BOOL ge;
  if(fpe) fpsetmask(fpe_flags);
  ge = (d >= [value get] + epsilon) ? YES : NO;
  if(fpe) fpsetmask(0);
  return ge;
}

-(BOOL)lt: (Double *)value {
  BOOL lt;
  if(fpe) fpsetmask(fpe_flags);
  lt = (d < [value get] + epsilon) ? YES : NO;
  if(fpe) fpsetmask(0);
  return lt;
}

@end
